/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <stdio.h>

#include "gtkeditor.h"
#include "internal.h"
#include "scanner.h"
#include "highlight.h"
#include "edit.h"

#define TOKEN_LOOKBACK 5

struct _GtkEditorBlockNode {
  GtkEditorBlock *block;
  gint block_id;
  gint offset;
  GtkEditorToken *end;
  GtkEditorBlockNode *next;
};

/*
 * globals
 */

/* pointer to current editor being analized */
static GtkEditor *g_editor;

static GtkEditorToken *g_start_token;
static int g_pos, g_length;

/* text changes associated with the newly inserted text */
static GSList *g_new_tags = NULL, *g_removal_tags = NULL;
static int g_blocks_changed;

static GtkEditorBlockNode *g_block_stack;
static GtkEditorBlockNode *g_block_removal_stack;

/* used for recyclying */
static GtkEditorToken *g_free_tokens = NULL;
static GtkEditorBlockNode *g_free_block_nodes = NULL;


/*
 * prototypes
 */
static void            move_position_near    (GtkEditor *editor,
                                              int pos);
static void            look_back             (GtkEditor *editor,
                                              int count);
static int             get_tag_id            (GtkEditor *editor,
                                              GtkEditorToken *token,
                                              int position,
                                              int length);
static void            split_token           (GtkEditorToken *token,
					      int              pos);
static GtkEditorToken *get_new_token         (void);
static void            recycle_token         (GtkEditorToken *token);
static GtkEditorBlock *get_new_block         (void);
static void            free_block            (GtkEditorBlock *block);
static GtkEditorBlockNode *
                       push_block            (GtkEditorBlockNode *block_stack,
                       			      GtkEditorBlock *block,
                       			      int block_id,
                       			      int offset,
                       			      GtkEditorToken *end);
static GtkEditorBlockNode *
		       pop_blocks            (GtkEditorBlockNode **block_stack,
		                              int id);
static GtkEditorBlockNode *
                       get_new_block_node   ();
static void            recycle_block_stack  (GtkEditorBlockNode **block_stack);
static void            push_blocks           (GtkEditorBlockNode **block_stack,
					      GtkEditorBlockNode *stack);
static GtkEditorBlock *pop_block             (GtkEditorBlockNode **block_stack);
static GtkEditorBlockNode *
                       duplicate_block_stack (GtkEditorBlockNode *block_stack);
static void            restore_block_stack   (GtkEditor *editor,
					      GtkEditorBlockNode *block_stack);
static void            update_blocks         (GtkEditor *editor,
	       				      GtkEditorToken *start,
	       				      int pos);


#if DEBUG
static void print_stack(GtkEditorBlockNode *stack);
static void print_tokens (GtkEditor *editor);
#endif

/* gtk_editor_insert_text

   This function is called when new text is inserted by the user. It
   scans the new text with the effected areas, and updates the editors
   token list.
*/
void
gtk_editor_insert_text(GtkTextBuffer *buffer, GtkTextIter *pos,
                       const gchar *text, gint new_text_length)
{
  GtkEditor *editor = GTK_EDITOR (buffer);
  GtkTextIter scan_pos, start, end;
  int i;
  int position = gtk_text_iter_get_char_index (pos) - new_text_length;

  if (!editor->scanner) {
    /* apply default tag */
    gtk_text_buffer_get_iter_at_char (buffer, &start, 0);
    gtk_text_buffer_get_iter_at_char (buffer, &end, new_text_length);
    gtk_text_buffer_apply_tag (buffer, "default", &start, &end);
    return;
  }
      
  g_insertion_point = position;
  g_insertion_length = new_text_length;
  g_length = gtk_text_buffer_get_char_count (buffer) - 1;

  if (!editor->tokens && g_length) {
    /* Uninitialized editor, just highlight the entire buffer.*/
    gtk_editor_hilite_buffer (editor);
    return;
  }

  /*
     Create a new token and insert it into the token list. This token
     represents the inserted text. A prop_id of -1 represents a token
     with no associated text in the editor. This fact is used in the
     merge routine below.
  */
  g_token = get_new_token();
  g_token->length = new_text_length;
  g_blocks_changed = FALSE;
  g_block_removal_stack = NULL;

  if (!editor->tokens) {
    /* empty editor */
    editor->tokens = g_token;
    g_start_token = NULL;
  } else {
    /* find insertion point */
    move_position_near(editor, position);
    /* insert the token into the token list */
    if (position == 0) {
      g_token->next = editor->tokens;
      editor->tokens->prev = g_token;
      editor->tokens = g_token;
      editor->cur_token_pos += new_text_length;
    } else if (position == (gtk_text_buffer_get_char_count(buffer)-1-new_text_length)) {
      g_token->prev = editor->cur_token;
      editor->cur_token->next = g_token;
    } else {
      if (editor->cur_token_pos != position) {
        if (editor->cur_token->block) {
          if (editor->cur_token->begins_block) {
            /* token needs to be split, destroy the block */
            g_block_removal_stack = push_block(g_block_removal_stack, NULL, editor->cur_token->block_id,
              editor->cur_token_pos, editor->cur_token->block->end);
            free_block(pop_block (&editor->block_stack));
            editor->cur_token->block = NULL;
            editor->cur_token->block_id = -1;
            editor->cur_token->begins_block = 0;
          } else {
            editor->block_stack->offset = editor->cur_token_pos;
            editor->cur_token->block->changed = TRUE;
            editor->cur_token->block->end = NULL;
            editor->cur_token->block = NULL;
          }
          g_blocks_changed = 1;
        }
        split_token(editor->cur_token, position - editor->cur_token_pos);
        editor->cur_token = editor->cur_token->next;
        editor->cur_token_pos = position + new_text_length;
      } else
        editor->cur_token_pos += new_text_length;
      g_token->next = editor->cur_token;
      g_token->prev = editor->cur_token->prev;
      editor->cur_token->prev->next = g_token;
      editor->cur_token->prev = g_token;
    }
    /* look back a certain ammount */
    look_back(editor, TOKEN_LOOKBACK);
    g_start_token = g_token = editor->cur_token;

    if (g_token->block && g_token->begins_block)
      pop_block (&editor->block_stack);
  }

  /* scan the text */
  g_editor = editor;
  editor->cur_token = NULL;
  g_removal_tags = NULL;
  g_new_tags = NULL;
  gtk_text_buffer_get_iter_at_char (buffer, &scan_pos, editor->cur_token_pos);
  _gtk_editor_scan (editor, &scan_pos, position, new_text_length);

  if(g_blocks_changed)
    update_blocks(editor, g_token, g_pos);
   
  gtk_text_buffer_get_iter_at_char (buffer, &start, position);
  gtk_text_buffer_get_iter_at_char (buffer, &end, position+new_text_length);
  
  for (i = 0; i < editor->n_tag_entries; i++) {
    if (editor->tag_entries[i]
        && gtk_text_iter_has_tag (&start, editor->tag_entries[i])) {
      gtk_text_buffer_remove_tag (buffer, editor->tag_entries[i]->name, &start, &end);
    }
  } 

  if (g_removal_tags)
    _gtk_editor_remove_tags (editor, g_removal_tags);

  if (g_new_tags)
    _gtk_editor_apply_tags (editor, g_new_tags);
        
  /* apply default tag */
  gtk_text_buffer_apply_tag (buffer, "default", &start, &end);

  restore_block_stack (editor, g_block_stack);
  /* print_tokens (editor); */
}

static int deletion_length;
static int empty_buffer_flag;

void
gtk_editor_delete_text_length (GtkTextBuffer *buffer,
                               GtkTextIter *start,
                               GtkTextIter *end)
{
  int start_pos, end_pos, t, length;
  if ((length = gtk_text_buffer_get_char_count(buffer)) == 1) {
    empty_buffer_flag = TRUE;
    return;
  }
  start_pos = gtk_text_iter_get_char_index(start);
  end_pos = gtk_text_iter_get_char_index(end);
  
  if (start_pos > end_pos) {
    t = start_pos;
    start_pos = end_pos;
    end_pos = t;
  }
  
  if (end_pos == length)
    end_pos -= 1;
  deletion_length = end_pos - start_pos;
  
  if (deletion_length == 0)
    gtk_signal_emit_stop_by_name    (GTK_OBJECT (buffer),
                                     "delete_text");
}


/* gtk_editor_delete_text

   This function is called when the user deletes text. It deletes the
   asociated tokens and block and rescans the affected area.

*/
void
gtk_editor_delete_text (GtkTextBuffer *buffer,
                        GtkTextIter *start,
                        GtkTextIter *end)
{
  GtkEditor *editor = GTK_EDITOR (buffer);
  GtkEditorToken *prev = NULL, *temp, *cur;
  GtkEditorBlockNode *node, *stack = NULL;
  GtkTextIter scan_pos;
  int start_pos = gtk_text_iter_get_char_index(start);
  int end_pos = start_pos + deletion_length;
  int pos, found;

  if (!editor->scanner) {
    /* No scanner, the text is already deleted, so return. */
    return;
  }
    
  if (empty_buffer_flag) {
    empty_buffer_flag = FALSE;
    return;
  }

  g_insertion_point = start_pos;
  g_block_removal_stack = NULL;
  g_blocks_changed = 0;

  /* find nearest token */
  move_position_near(editor, start_pos);

  if (start_pos > editor->cur_token_pos) {
    /* starts in the middle of a token */
    if (editor->cur_token->block) {
      if (editor->cur_token->begins_block) {
        /* token needs to be split, destroy the block */
        g_block_removal_stack = push_block(g_block_removal_stack, NULL, editor->cur_token->block_id,
          editor->cur_token_pos, editor->cur_token->block->end);
        free_block(pop_block (&editor->block_stack));
        editor->cur_token->block = NULL;
        editor->cur_token->block_id = -1;
        editor->cur_token->begins_block = 0;
      } else {
        editor->block_stack->offset = start_pos;
        editor->cur_token->block->changed = TRUE;
        editor->cur_token->block->end = NULL;
        editor->cur_token->block = NULL;
      }
      g_blocks_changed = 1;
    }

    /* split the token */
    split_token (editor->cur_token, start_pos - editor->cur_token_pos);
    editor->cur_token = editor->cur_token->next;
    editor->cur_token_pos = start_pos;
  } else if (editor->cur_token->block && editor->cur_token->begins_block)
    pop_block (&editor->block_stack);

  cur = editor->cur_token;

  /* set prev to the token prior to the deleted area */
  if (cur->prev)
    prev = cur->prev;

  /* get rid of tokens and blocks in the deleted area */
  pos = start_pos;
  while (pos < end_pos) {
    if (cur->block) {
      if(cur->begins_block) {
        g_block_removal_stack = push_block(g_block_removal_stack, NULL, cur->block_id,
                   start_pos, cur->block->end);
        free_block(cur->block);
        cur->block = NULL;
        cur->block_id = -1;
        cur->begins_block = 0;
      } else {
        found = 0;
        for (node = g_block_removal_stack; node != NULL; node = node->next) {
          if(node->end == cur) {
            stack = node;
            found = 1;
          }
        }

        if (found) {
          node = g_block_removal_stack;
          g_block_removal_stack = stack->next;
          stack->next = NULL;
          recycle_block_stack(&node);
        }

        for (node = editor->block_stack; node != NULL; node = node->next) {
          if(node->end == cur) {
            node->offset = start_pos;
            cur->block->changed = TRUE;
            node->block->end = NULL;
          }
        }
      }
      g_blocks_changed = 1;
    }

    if (pos + cur->length <= end_pos) {
      /* Token is entirely inside the deleted area, recycle it. */
      pos += cur->length;
      temp = cur;
      cur = cur->next;
      recycle_token(temp);
    } else {
      /* Token covers more than the deleted area, shorten it. */
      cur->length -= end_pos - pos;
      pos = end_pos;
    }
  }

  if((gtk_text_buffer_get_char_count(buffer) - 1) == 0) {
    /* cleared editor, get rid of everything */
    editor->tokens = NULL;
    editor->cur_token = NULL;
    editor->block_stack = NULL;
    return;
  }

  if (prev) {
    prev->next = cur;
    if (cur)
      cur->prev = prev;
    editor->cur_token = prev;
    editor->cur_token_pos = start_pos - prev->length;
    /* check for crossing block boundries */
    if (prev->block && prev->begins_block == FALSE) {
      push_blocks(&editor->block_stack, prev->block->stack);
      prev->block->stack = NULL;
    }
    /* look back a certain ammount */
    look_back(editor, TOKEN_LOOKBACK);
  } else {
    cur->prev = NULL;
    editor->tokens = cur;
    editor->cur_token = cur;
    editor->cur_token_pos = start_pos;
  }

  g_start_token = g_token = editor->cur_token;

  if (g_token->block && g_token->begins_block)
    pop_block (&editor->block_stack);

  /* rescan affected text */
  g_editor = editor;
  editor->cur_token = NULL;
  g_removal_tags = NULL;
  g_new_tags = NULL;
  gtk_text_buffer_get_iter_at_char (buffer, &scan_pos, editor->cur_token_pos);
  _gtk_editor_scan (editor, &scan_pos, start_pos, 0);

  if(g_blocks_changed)
    update_blocks(editor, g_token, g_pos);
		
  if (g_removal_tags)
    _gtk_editor_remove_tags (editor, g_removal_tags);

  if (g_new_tags)
    _gtk_editor_apply_tags (editor, g_new_tags);

  restore_block_stack (editor, g_block_stack);
}

/* gtk_editor_hilite_buffer

   Rescans the entire buffer and highlights acordingly.
*/
void
gtk_editor_hilite_buffer (GtkEditor *editor)
{
  int length;
  GtkTextBuffer *buffer;
  GtkEditorToken *cur, *temp;
  GtkTextIter scan_pos, start, end;

  if (editor->scanner == NULL)
    /* nothing to do here */
    return;

  buffer = GTK_TEXT_BUFFER (editor);

  if (!(length = gtk_text_buffer_get_char_count (buffer)-1))
    return;
  if (editor->tokens) {
    for(cur = editor->tokens; cur != NULL;) {
      temp = cur;
      cur = cur->next;
      recycle_token (temp);
    }
  }

  /*
     Create a new token an insert it into the token list. This token
     represents the entire buffer. A prop_id of -1 represents a token
     with no associated text in the editor. This fact is used in the
     merge routine below.
  */
  g_token = editor->tokens = get_new_token();
  editor->cur_token = NULL;
  g_token->length = length;
  g_start_token = NULL;
  g_new_tags = NULL;
  g_block_removal_stack = NULL;
  g_editor = editor;
  gtk_text_buffer_get_iter_at_char (buffer, &scan_pos, 0);
  _gtk_editor_scan (editor, &scan_pos, 0, length);

  if (g_new_tags)
    _gtk_editor_apply_tags (editor, g_new_tags);
     
  /* apply default tag */
  gtk_text_buffer_get_iter_at_char (buffer, &start, 0);
  gtk_text_buffer_get_iter_at_char (buffer, &end, length);
  gtk_text_buffer_apply_tag (buffer, "default", &start, &end);

  restore_block_stack (editor, g_block_stack);
}

/* move_position_near

   Moves the the editors current token near pos.

*/
static void
move_position_near(GtkEditor *editor, int pos)
{
  GtkEditorToken *token = editor->cur_token;
  int cur_pos = editor->cur_token_pos;

  if (pos >= cur_pos + token->length) {
    /* Position is greater than the current position, search forward. */
    while (cur_pos + token->length <= pos && token->next != NULL) {
      /* increment the token */
      cur_pos += token->length;
      token = token->next;

      /* check for crossing block boundries */
      if (token->prev && token->prev->block
          && token->prev->begins_block == FALSE) {
        push_blocks(&token->prev->block->stack,
          pop_blocks(&editor->block_stack, token->prev->block_id));
      }
      if (token->block && token->begins_block) {
        editor->block_stack = push_block(editor->block_stack, token->block,
                   token->block_id, cur_pos, token->block->end);
      }
    }
  } else if(pos < cur_pos) {
    /* Position is less than the current position, search back */
    while (cur_pos > pos) {
      /* decrement the token */
      token = token->prev;
      cur_pos -= token->length;
      /* check for crossing block boundries */
      if (token->next->block && token->next->begins_block)
        pop_block(&editor->block_stack);
      if (token->block && token->begins_block == FALSE) {
        push_blocks(&editor->block_stack, token->block->stack);
        token->block->stack = NULL;
      }
    }
  }

  editor->cur_token = token;
  editor->cur_token_pos = cur_pos;
}

static void
look_back (GtkEditor *editor, int count)
{
  GtkEditorToken *token = editor->cur_token;
  int pos = editor->cur_token_pos;

  /* look back */
  while (token->prev != NULL && count > 0) {
    count--;
    /* decrement the token */
    token = token->prev;
    pos -= token->length;
    /* check for crossing block boundries */
    if (token->next->block && token->next->begins_block)
      pop_block(&editor->block_stack);
    if (token->block && token->begins_block == FALSE) {
      push_blocks(&editor->block_stack, token->block->stack);
      token->block->stack = NULL;
    }
  }
  editor->cur_token = token;
  editor->cur_token_pos = pos;
}

/* merge_token

   This function merges a token into the token list. The merged token is
   returned.
*/
void
_gtk_editor_merge_token (int id, int precedence, int length, int position,
                         int block_id, int nesting, int begins_block, int state)
{
  GtkEditorToken *temp, *cur = NULL;
  GtkEditorBlockNode *stack = NULL, *node, *tnode;
  unsigned int pos = 0, len, cur_pos = 0, from, to;
  int tag_id = -1, found, flag;

  g_pos = position;
  /* Check for a previous end of block */
  if (g_token != g_start_token
      && g_token->prev && g_token->prev->block
      && g_token->prev->begins_block == FALSE) {
    /* Here we have crossed a block end, we must pop all the blocks
       contained in this block off of the highlight stack and push
       them onto this block.
    */
    stack = pop_blocks(&g_editor->block_stack, g_token->prev->block_id);
    if (!g_editor->cur_token)
      push_blocks(&g_token->prev->block->stack, stack);
  }
  
  /* This code below handles when a block token has changed */
  if (g_token->block_id != -1 && g_token->block
      && (g_token->block_id != block_id || g_token->begins_block != begins_block)) {
    if (g_token->begins_block) {
      g_block_removal_stack = push_block(g_block_removal_stack, NULL, g_token->block_id,
                 g_pos, g_token->block->end);
      free_block(g_token->block);
      g_token->block = NULL;
    } else {
      found = 0;
      for (node = g_block_removal_stack; node != NULL; node = node->next) {
        if(node->end == g_token) {
          flag = 0;
          for (tnode = g_editor->block_stack; tnode != NULL; tnode = tnode->next) {
            if(tnode->block_id == node->block_id) {
              flag = 1;
              break;
            }
          }
                    
                          
          if (flag) {
            if (g_token->block_id == node->block_id)
              g_token->block = NULL;                
          } else if (g_token->block_id == node->block_id) {
            if (g_editor->tag_entries[node->block_id]) {
              g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                             node->block_id,
        				                       node->offset,
        				                       g_pos+g_token->length);
            }
            g_token->block = NULL;
          } else if (g_editor->tag_entries[node->block_id]) {
            g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                           node->block_id,
        				                     node->offset,
        				                     g_pos);
          }
          stack = node;
          found = 1;
        }
      }

      if (found) {
        node = g_block_removal_stack;
        g_block_removal_stack = stack->next;
        stack->next = g_free_block_nodes;
        g_free_block_nodes = node;
      }

      for (node = g_editor->block_stack; node != NULL; node = node->next) {
        if(node->end == g_token) {
          if (node->block_id == g_token->block_id)
            node->offset = g_pos+g_token->length;
          else
            node->offset = g_pos;                   
          g_token->block->changed = TRUE;
          g_token->block = NULL;
          node->block->end = NULL;
          g_blocks_changed = TRUE;
        }
      }
      
    } 
  }
    
  g_token->id = id;
  g_token->block_id = block_id;
  g_token->begins_block = begins_block;
  g_token->precedence = precedence;
  g_token->nest = nesting;
  g_token->state = state;

  tag_id = get_tag_id(g_editor, g_token, position, length);

  if (g_insertion_length
      && ((g_pos >= g_insertion_point && g_pos < g_insertion_point+g_insertion_length)
          || (g_insertion_point >= g_pos && g_insertion_point < g_pos+length)))  {
    if (g_pos < g_insertion_point)
      from = g_insertion_point;
    else
      from = g_pos;
    
    if (g_pos + length > g_insertion_point + g_insertion_length)
      to = g_insertion_point + g_insertion_length;
    else
      to = g_pos + length;            
                        
                                                
    for (node = g_editor->block_stack; node != NULL; node = node->next) {
      if (g_editor->tag_entries[node->block_id])
          g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
						   node->block_id,
						   from,
						   to);
    }
  }

  /* Loop through the tokens that intersect the merged token. */
  while (pos < length) {
    /* block stuff */
    if(cur != NULL && g_token->block) {
      if(g_token->begins_block) {
        g_block_removal_stack = push_block(g_block_removal_stack, NULL, g_token->block_id,
                   g_pos+pos, g_token->block->end);
        free_block(g_token->block);
        g_token->block = NULL;
      } else {
        found = 0;
        for (node = g_block_removal_stack; node != NULL; node = node->next) {
          if(node->end == g_token) {
            if (g_token->block_id == node->block_id) {
              if (g_editor->tag_entries[node->block_id]) {
                g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                             node->block_id,
        				                       node->offset,
        				                       position+pos+g_token->length);
              }
              g_token->block = NULL;
            } else if (g_editor->tag_entries[node->block_id]) {
              g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                           node->block_id,
        				                     node->offset,
        				                     position+pos);
            }
            stack = node;
            found = 1;
          }
        }

        if (found) {
          node = g_block_removal_stack;
          g_block_removal_stack = stack->next;
          stack->next = g_free_block_nodes;
          g_free_block_nodes = node;
        }

        for (node = g_editor->block_stack; node != NULL; node = node->next) {
          if(node->end == g_token) {
            if (node->block_id == g_token->block_id)
              node->offset = g_pos+pos+g_token->length;
            else
              node->offset = g_pos+pos;  
            g_token->block->changed = TRUE;
            node->block->end = NULL;
          }
        }
      }
      g_blocks_changed = 1;
    }

    /* actual merging */
    if (length - pos >= g_token->length) {
      /* new token overlaps the current token */
      len = g_token->length;

      if (g_token->tag_id != tag_id) {
        if (g_token->tag_id != -1)
          g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
						       g_token->tag_id,
						       g_pos+pos,
						       g_pos+pos+len);
	if (tag_id != -1)
          g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
						   tag_id,
						   g_pos+pos,
						   g_pos+pos+len);
      }

      if (cur == NULL) {
        /* Then merged token doesn't belong to the list yet. Since we would be
           getting rid of this token normally, we can let the merged
           token take over it's spot.
        */
        cur = g_token;
	cur_pos = g_pos;
	g_token = g_token->next;
      } else {
        /* The token no longer exist, recycle it. */
        temp = g_token;
	if (g_token->prev)
	  g_token->prev->next = g_token->next;
	if (g_token->next)
	  g_token->next->prev = g_token->prev;
	g_token = g_token->next;
	recycle_token (temp);
      }

    } else {
      /* The merged token covers less than the current token. */
      len = length - pos;

      if (g_token->tag_id != tag_id) {
        if (g_token->tag_id != -1)
          g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
						       g_token->tag_id,
						       g_pos+pos,
						       g_pos+pos+len);
	if (tag_id != -1)
          g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
						   tag_id,
						   g_pos+pos,
						   g_pos+pos+len);
      }
		
      if (cur == NULL) {
        /* Then merged token doesn't belong to the list yet. Since the merged
           token covers less than this token we have to split it. */
	cur = g_token;
	cur_pos = g_pos;
	split_token(g_token, len);
	g_token = g_token->next;
      } else {
	/* shorten token */
	g_token->length -= len;
      }
    }
    /* increment the ammount merged */
    pos += len;
  }

  cur->tag_id = tag_id;
  cur->length = length;

  /* update the cur token */
  if ((g_insertion_point >= g_pos
      && g_insertion_point < g_pos + length) ||
      (cur->next == NULL && g_insertion_point == g_pos + length)) {
    g_editor->cur_token = cur;
    g_editor->cur_token_pos = g_pos;
    g_block_stack = duplicate_block_stack(g_editor->block_stack);
  }

  g_pos += length;
}

static void
update_blocks (GtkEditor *editor,
	       GtkEditorToken *start,
	       int pos)
{
  GtkEditorToken *token, *prev = NULL;
  GtkEditorBlockNode *stack;
  int tag_id;

  for (token = start; token != NULL; token = token->next) {
    /* Check for a previous end of block */
    if (token->prev && token->prev->block
        && token->prev->begins_block == FALSE) {
      /* Here we have crossed a block end, we must pop all the blocks
         contained in this block off of the highlight stack and push
         them onto this block. */
      stack = pop_blocks(&editor->block_stack, token->prev->block_id);
      recycle_block_stack(&stack);
    }

    tag_id = get_tag_id (editor, token, pos, token->length);

    if (token->tag_id != tag_id) {
      if (token->tag_id != -1)
        g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
						     token->tag_id,
						     pos, pos+token->length);
      token->tag_id = tag_id;
      if (tag_id != -1)
        g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
						 tag_id,
						 pos, pos+token->length);
    }

    pos += token->length;
    prev = token;
  }

  if (g_block_removal_stack) {
    /* we have some unterminated blocks, lets finish them */
    for (stack = g_block_removal_stack; stack != NULL; stack = stack->next) {
      if (editor->tag_entries[stack->block_id]) {
        g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                     stack->block_id,
        				             stack->offset,
        				             pos);
      }
    }
    recycle_block_stack(&g_block_removal_stack);
  }

  if (editor->block_stack) {
    /* we have some unterminated blocks, lets finish them */
    if(prev != NULL && editor->block_stack->block->end == prev) {
      /* this is already taken care of below so we have to pop it */
      pop_block (&editor->block_stack);
    }

    for (stack = editor->block_stack; stack != NULL; stack = stack->next) {
      if (stack->block->changed && editor->tag_entries[stack->block_id]) {
        g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
                                                 stack->block_id,
        				         stack->offset,
        				         pos);
        stack->block->changed = FALSE;
        stack->block->end = NULL;
      }
      stack->end = NULL;
    }
  }
}

/* get_tag_id

   This function returns the tag id for the token passed in. A tag is a group
   of properties that can be ascociated with a group of text. In addition
   to getting the tag id it also updates the block stacks.
   
   FIXME: this function is a complete mess, there is a lot of hacks currently,
          the logic should be cleaned up a lot.
*/

static int
get_tag_id(GtkEditor *editor, GtkEditorToken *token, int position, int length)
{
  GtkEditorBlockNode *node, *cur = NULL, *tnode;
  int found = 0, tag_id = -1, flag;

  /* Check for the begining of a block*/
  if (token->block_id != -1 && token->begins_block &&
      ((editor->block_stack && editor->block_stack->block->begin->nest
        && token->precedence >= editor->block_stack->block->begin->precedence)
       || !editor->block_stack)) {
    /* Here we have encountered a legal begin block.
       We can push it on to the highlight stack and start highlighting
       as this type block. */

    if (!token->block) {
      /* This block didn't exist before, we must create a new block. */
      g_blocks_changed = 1;
      token->block = get_new_block();
      token->block->begin = token;
      token->block->end = NULL;
      token->block->changed = TRUE;
      /* Push the block onto the stack, and indicate it is a new block */
      editor->block_stack = push_block (editor->block_stack, token->block, token->block_id, position, NULL);
    } else {
      /* This block already exist, push it on to stack */
      editor->block_stack = push_block (editor->block_stack, token->block, token->block_id, position, token->block->end);
    }

    /* If the block begin has a token, it takes precedence over the block tag */
    if(token->id != -1 && editor->tag_entries[token->id] != NULL)
      tag_id = token->id;
  } else if (token->block_id != -1 && token->begins_block == FALSE){
    found = 0;
    /* This token is the end of a block, check if ends anything that came before it. */
    for (node = g_block_removal_stack; node != NULL; node = node->next) {
      if(node->end == token) {
        flag = 0;
        for (tnode = g_editor->block_stack; tnode != NULL; tnode = tnode->next) {
          if(tnode->block_id == node->block_id && !tnode->block->changed) {
            flag = 1;
            break;
          }
        }
                    
                          
        if (flag) {
          if (token->block_id != node->block_id)
            token->block = NULL;                         
        } else if (token->block_id != node->block_id) {
          if (editor->tag_entries[token->block_id]) {
            g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                         g_token->block_id,
        				                 node->offset,
        				                 position);
          }
        } else if (editor->tag_entries[node->block_id]) {
          g_removal_tags = _gtk_editor_add_tag_change (g_removal_tags,
                                                       node->block_id,
        				               node->offset,
        				               position+length);
        	 token->block = NULL;
        }
        cur = node;
        found = 1;
      }
    }

    if (found) {
      node = g_block_removal_stack;
      g_block_removal_stack = cur->next;
      cur->next = g_free_block_nodes;
      g_free_block_nodes = node;
    }

    for (node = editor->block_stack; node != NULL; node = node->next) {
      if(token->block_id == node->block_id) {
        for (cur = editor->block_stack; cur != node; cur = cur->next) {
          if (cur->block->changed) {
            if(editor->tag_entries[cur->block_id]) {
              g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
                                                       cur->block_id,
        				               cur->offset,
        				               position);
            }
            cur->block->changed = FALSE;
          } else if (cur->end != token) {
            g_block_removal_stack = push_block (g_block_removal_stack, NULL, cur->block_id, position, cur->end);
          }
          cur->block->end = token;
          cur->end = token;
        }
        if (node->block->changed) {
          int l;
          l = g_insertion_point + g_insertion_length;
          if (!(l >= position + length) 
              && (editor->tag_entries[token->block_id])) {
            g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
                                                     node->block_id,
        				             node->offset,
        				             position+length);
          }
          node->block->changed = FALSE;
        } else if (node->block->end != token) {
          if (g_insertion_point <= position + length
              && g_insertion_point + g_insertion_length == g_length
              && editor->tag_entries[token->block_id])
            g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
                                                     node->block_id,
        				             g_insertion_point,
        				             position+length);
          else {
            flag = 0;
            for (tnode = g_editor->block_stack; tnode != NULL; tnode = tnode->next) {
              if(tnode->block_id == node->block_id && !tnode->block->changed) {
                flag++;
              }
            }
            if (flag == 1)
              g_block_removal_stack = push_block (g_block_removal_stack, NULL, node->block_id, position+length, node->block->end);
          }
          g_blocks_changed = 1;
        } else if (g_block_removal_stack) {
          for (cur = g_block_removal_stack; cur != NULL; cur = cur->next) {
            if (cur->block_id == node->block_id) {
              /* This fixes the problem of a nested block being removed
                 that strips this block of it's properties */
              g_new_tags = _gtk_editor_add_tag_change (g_new_tags,
                                                       node->block_id,
        				               node->offset,
        				               position+length);
            }
          }
        }
        token->block = node->block;
        node->block->end = token;
        node->end = token;             
        if (token->id != -1 && editor->tag_entries[token->id] != NULL)
          tag_id = token->id;
        found = 1;
        break;
      }
    }

    if (!found)
      token->block = NULL;

  } else if (token->block && token->begins_block) {
    g_blocks_changed = 1;
    g_block_removal_stack = push_block (g_block_removal_stack, NULL, token->block_id, position, token->block->end);
    free_block(token->block);
    token->block = NULL;
  }

  if (tag_id == -1) {
    if (editor->block_stack) {
      if (token->precedence > editor->block_stack->block->begin->precedence && token->id != -1
          && editor->tag_entries[token->id] != NULL) {
        tag_id = token->id;
      }
    } else if (token->id != -1 && editor->tag_entries[token->id] != NULL) {
      tag_id = token->id;
    }
  }

  return tag_id;
}

static void
split_token (GtkEditorToken *token, int pos)
{
  GtkEditorToken *new_token;
  
  new_token = get_new_token();
  new_token->next = token->next;
  new_token->prev = token;
  if (token->next)
    token->next->prev = new_token;
  token->next = new_token;
  new_token->length = token->length - pos;
  new_token->tag_id = token->tag_id;
  new_token->state = token->state;
  token->length = pos;
}

static GtkEditorToken*
get_new_token (void)
{
  static GtkEditorToken *temp;
  
  if (g_free_tokens) {
    temp = g_free_tokens;
    g_free_tokens = g_free_tokens->next;
    if (g_free_tokens)
      g_free_tokens->prev = NULL;
  } else {
    temp = g_new (GtkEditorToken, 1);
  }
  temp->id = -1;
  temp->block_id = -1;
  temp->precedence = 0;
  temp->tag_id = -1;
  temp->length = 0;
  temp->nest = 0;
  temp->state = 0;
  temp->block = NULL;
  temp->next = NULL;
  temp->prev = NULL;

  return temp;
}

static void
recycle_token(GtkEditorToken *token)
{
  token->prev = NULL;
  token->next = g_free_tokens;
  if (g_free_tokens)
    g_free_tokens->prev = token;
  g_free_tokens = token;
}

static GtkEditorBlockNode *
push_block (GtkEditorBlockNode *block_stack, GtkEditorBlock *block,
            int block_id, int offset, GtkEditorToken *end)
{
  GtkEditorBlockNode *node;

  node = get_new_block_node ();
  node->block = block;
  node->block_id = block_id;
  node->offset = offset;
  node->end = end;
  node->next = block_stack;
  return node;
}

static void
push_blocks(GtkEditorBlockNode **block_stack, GtkEditorBlockNode *block_node)
{
  GtkEditorBlockNode *node;

  for(node = block_node; node->next != NULL; node = node->next);

  node->next = *block_stack;
  *block_stack = block_node;
}

static GtkEditorBlock *
pop_block (GtkEditorBlockNode **block_stack)
{
  GtkEditorBlock *block = NULL;
  GtkEditorBlockNode *temp;

  if (*block_stack) {
    block = (*block_stack)->block;
    temp = *block_stack;
    *block_stack = (*block_stack)->next;
    temp->next = NULL;
    recycle_block_stack (&temp);
  }

  return block;
}

static GtkEditorBlockNode *
pop_blocks (GtkEditorBlockNode **block_stack, int id)
{
  GtkEditorBlockNode *node, *stack = NULL;

  /* first we check if if this token ends any of the blocks on the stack */
  for (node = *block_stack; node != NULL; node = node->next) {
    if (node->block_id == id) {
      stack = *block_stack;
      *block_stack = node->next;
      node->next = NULL;
      break;
    }
  }

  return stack;
}

static GtkEditorBlockNode *
get_new_block_node ()
{
  GtkEditorBlockNode *node;

  if (g_free_block_nodes) {
    node = g_free_block_nodes;
    g_free_block_nodes = g_free_block_nodes->next;
  } else {
    node = g_new (GtkEditorBlockNode, 1);
  }

  node->block = NULL;
  node->block_id = -1;
  node->offset = 0;
  node->next = NULL;

  return node;
}

static void
recycle_block_stack (GtkEditorBlockNode **block_stack)
{
  GtkEditorBlockNode *node;

  if (*block_stack == NULL)
    return;

  for (node = *block_stack; node->next != NULL; node = node->next);

  node->next = g_free_block_nodes;
  g_free_block_nodes = *block_stack;
  *block_stack = NULL;
}

static GtkEditorBlockNode *
duplicate_block_stack (GtkEditorBlockNode *block_stack)
{
  GtkEditorBlockNode *new_block_stack = NULL, *node = NULL;

  while (block_stack != NULL) {
    if(new_block_stack == NULL)
      node = new_block_stack = get_new_block_node ();
    else
      node = node->next = get_new_block_node ();
    node->block = block_stack->block;
    node->block_id = block_stack->block_id;
    node->offset = block_stack->offset;
    node->end = block_stack->end;
    block_stack = block_stack->next;
  }

  if (node)
    node->next = NULL;

  return new_block_stack;
}

static void
restore_block_stack (GtkEditor *editor, GtkEditorBlockNode *block_stack)
{
  if (editor->block_stack)
    recycle_block_stack(&editor->block_stack);
  editor->block_stack = block_stack;
}

static GtkEditorBlock *
get_new_block (void)
{
  GtkEditorBlock *block;

  block = g_new(GtkEditorBlock, 1);
  block->begin = NULL;
  block->end = NULL;
  block->changed = FALSE;
  block->stack = NULL;

  return block;
}

static void
free_block (GtkEditorBlock *block)
{
  g_free(block);
}

/*
  for debugging
*/
#if DEBUG
static void
print_stack(GtkEditorBlockNode *stack)
{
  g_print("Stack state:\n");
  for (; stack != NULL; stack = stack->next) {
    g_print("%d\n", stack->block_id);
  }
}

static void
print_tokens (GtkEditor *editor)
{
  GtkEditorToken *token = editor->tokens;
  GtkTextIter iter;
  int i, pos = 0, level = 0;

  g_print("--------------------------------------------\n");
  print_stack(editor->block_stack);

  gtk_text_buffer_get_iter_at_char(GTK_TEXT_BUFFER(editor), &iter, 0);
  g_print("\n");
  for (; token != NULL; token = token->next) {
    g_print("|");
    if(token == editor->cur_token)
      g_print("*->|");

    if (token->block && token->begins_block) {
      g_print("%d>|", level++);
    }
    for(i = 0; i < token->length; i++) {
      g_print("%c", gtk_text_iter_get_char(&iter));
      if(!gtk_text_iter_forward_char (&iter))
        break;
    }
    if(token->block && token->begins_block == FALSE) {
      g_print("|<%d", --level);
    }
    pos += token->length;
  }
  g_print("|\n");
  g_print("--------------------------------------------\n");
}

#endif
