/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <stdio.h>

#include "internal.h"
#include "edit.h"
#include "highlight.h"

typedef struct _TagChange {
  int tag_id;
  unsigned int start;
  unsigned int end;
}TagChange;

/*
 * globals
 */
/* for recycling */
static GSList *g_free_tag_changes = NULL;

GtkTextTag *
gtk_editor_get_default_tag (GtkEditor *editor)
{
  return editor->default_tag;
}

GtkTextTag *
gtk_editor_create_tag (GtkEditor *editor, char *pattern_name)
{
  int id;

  if (!editor->scanner)
    return NULL;

  id = editor->scanner->get_token_no (pattern_name);
  if (id == -1)
    id = editor->scanner->get_block_no (pattern_name);
  if (id != -1) {
    if(editor->tag_entries[id] == NULL)
      editor->tag_entries[id] =
        gtk_text_buffer_create_tag (GTK_TEXT_BUFFER(editor),
                                    pattern_name);
    return editor->tag_entries[id];
  }

  return NULL;
}

GSList *
_gtk_editor_add_tag_change (GSList *tag_changes,
			    int tag_id, unsigned int start,
			    unsigned int end)
{
  GSList *cur;
  TagChange *prop;
    
  if (start == end)
    return tag_changes;
  
  /* check for overaping tags, this kind of a lazy hack for now */ 
  for (cur = tag_changes; cur != NULL; cur = cur->next) {
    prop = cur->data;
    if (prop->tag_id == tag_id
        && ((start >= prop->start && start < prop->end)
            || (prop->start >= start && prop->start < end))) {
      prop->start = (prop->start < start)? prop->start : start;
      prop->end = (prop->end > end)? prop->end : end;
      return tag_changes; 
    }
  }

  if (g_free_tag_changes) {
    cur = g_free_tag_changes;
    g_free_tag_changes = g_slist_remove_link (g_free_tag_changes, cur);
    prop = cur->data;
  } else {
    prop = g_new (TagChange, 1);
    cur = g_slist_alloc();
    cur->data = prop;
    cur->next = NULL;
  }

  prop->tag_id = tag_id;
  prop->start = start;
  prop->end = end;

  if(tag_changes == NULL)
    return cur;
  return(g_slist_concat(cur, tag_changes));
}

void
_gtk_editor_apply_tags (GtkEditor *editor,
			GSList *tag_changes)
{
  GSList *cur;
  TagChange *prop;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(editor);
  GtkTextIter start, end;

  if(tag_changes == NULL)
    return;
        
  tag_changes = g_slist_reverse(tag_changes);
  /*g_print("applying tags:\n");*/
  for(cur = tag_changes; cur != NULL;) {
    prop = cur->data;
    /*g_print("highlighting %d from %d to %d\n", prop->tag_id, prop->start, prop->end);*/
    gtk_text_buffer_get_iter_at_char (buffer, &start, prop->start);
    gtk_text_buffer_get_iter_at_char (buffer, &end, prop->end);
    gtk_text_buffer_apply_tag (buffer, editor->tag_entries[prop->tag_id]->name, &start, &end);
    tag_changes = g_slist_remove_link(tag_changes, cur);
    g_free_tag_changes = g_slist_concat(cur, g_free_tag_changes);
    cur = tag_changes;
  }
}

void
_gtk_editor_remove_tags (GtkEditor *editor,
			 GSList *tag_changes)
{
  GSList *cur;
  TagChange *prop;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(editor);
  GtkTextIter start, end;

  if(tag_changes == NULL)
    return;

  tag_changes = g_slist_reverse(tag_changes);
  /* g_print("removing tags:\n");*/
  for(cur = tag_changes; cur != NULL;) {
    prop = cur->data;
    /*g_print("unhighlighting %d from %d to %d\n", prop->tag_id, prop->start, prop->end);*/
    gtk_text_buffer_get_iter_at_char (buffer, &start, prop->start);
    gtk_text_buffer_get_iter_at_char (buffer, &end, prop->end);
    gtk_text_buffer_remove_tag (buffer, editor->tag_entries[prop->tag_id]->name, &start, &end);
    tag_changes = g_slist_remove_link(tag_changes, cur);
    g_free_tag_changes = g_slist_concat(cur, g_free_tag_changes);
    cur = tag_changes;
  }
}
