/*  Test of the gtk editor widget.
 *  Copyright (C) 1998, 2000 Thomas Mailund <mailund@daimi.au.dk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include "gtkeditor.h"
#include <string.h>

#define BUF_MAX_SIZE 128	/* seems fair for one line */

/* global vars */
static GtkEditor *mainbuffer;
static GtkNotebook *book;
static GtkEntry *minibuffer;
static void (*minibuffer_callback)(char *) = NULL;
static GtkEditorScanner *g_cscanner = NULL;

/* minibuffer stuff */
static gint
minibuf_key_press (GtkWidget *widget, GdkEventKey *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->keyval == GDK_Return) {
    gtk_widget_grab_focus (GTK_WIDGET (mainbuffer));
    if (minibuffer_callback) {
      /* The callback is called with a copy of the text. It needs to
       * take care of freeing the text, should that be done. */
      minibuffer_callback (g_strdup(gtk_entry_get_text (minibuffer)));
    }
    gtk_entry_set_text (minibuffer, "");
  }

  return FALSE;
}

static void
get_text_from_minibuffer (GtkSignalFunc callback)
{
  gtk_entry_set_text (minibuffer, "");
  minibuffer_callback = callback;
  gtk_widget_grab_focus (GTK_WIDGET (minibuffer));
}

/* callbacks */
/* callbacks */
static void
load_file_cb (char *filename)
{
  char buf[BUF_MAX_SIZE];
  FILE *fp = fopen (filename, "r");

#ifdef DEBUG
  g_print ("load_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  /* clear old text */
  if (gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(mainbuffer))) {
    gtk_text_buffer_delete_chars (GTK_TEXT_BUFFER(mainbuffer), 0, -1);
  }

  /* load new */
  do {
    if (fgets (buf, BUF_MAX_SIZE, fp) == NULL) {
      break;
    } else {
      gtk_text_buffer_insert_at_char (GTK_TEXT_BUFFER(mainbuffer),
                                      0, buf, strlen (buf));
    }
  } while (!feof (fp));

  /* clean up */
  fclose (fp);
  g_free (filename);
}

static void
save_file_cb (char *filename)
{
  int i, len;
  char *buf;
  FILE *fp = fopen (filename, "w");

#ifdef DEBUG
  g_print ("save_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  /* get_chars allocs a new buffer each time, so I have to do all this
   * g_free stuff */
  len = gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER (mainbuffer));
  for (i = 0; i < len - BUF_MAX_SIZE;) {
    buf = gtk_text_buffer_get_text_chars (GTK_TEXT_BUFFER (mainbuffer), i, i + BUF_MAX_SIZE, FALSE);
    fputs (buf, fp);
    i += BUF_MAX_SIZE;
    g_free (buf);
  }
  /* and the leftovers */
  buf = gtk_text_buffer_get_text_chars (GTK_TEXT_BUFFER (mainbuffer), i, len, FALSE);
  fputs (buf, fp);
  g_free (buf);

  fclose (fp);
}


static void
clear_buffer (void)
{
  gtk_text_buffer_delete_chars (GTK_TEXT_BUFFER(mainbuffer), 0, -1);
}

static void
load_file (void)
{
  get_text_from_minibuffer (load_file_cb);
}

static void
save_file (void)
{
  get_text_from_minibuffer (save_file_cb);
}

static gint
tag_event_handler(GtkTextTag *tag, GtkWidget *widget, GdkEvent *event,
                  const GtkTextIter *iter, gpointer user_data)
{
  gint char_index;

  char_index = gtk_text_iter_get_char_index(iter);

  switch (event->type)
    {
    case GDK_MOTION_NOTIFY:
      printf("Motion event at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_BUTTON_PRESS:
      printf("Button press at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_2BUTTON_PRESS:
      printf("Double click at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_3BUTTON_PRESS:
      printf("Triple click at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_BUTTON_RELEASE:
      printf("Button release at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_KEY_PRESS:
    case GDK_KEY_RELEASE:
    case GDK_ENTER_NOTIFY:
    case GDK_LEAVE_NOTIFY:
    case GDK_PROPERTY_NOTIFY:
    case GDK_SELECTION_CLEAR:
    case GDK_SELECTION_REQUEST:
    case GDK_SELECTION_NOTIFY:
    case GDK_PROXIMITY_IN:
    case GDK_PROXIMITY_OUT:
    case GDK_DRAG_ENTER:
    case GDK_DRAG_LEAVE:
    case GDK_DRAG_MOTION:
    case GDK_DRAG_STATUS:
    case GDK_DROP_START:
    case GDK_DROP_FINISHED:
    default:
      break;
    }

  return FALSE;
}

static void
setup_tag(GtkTextTag *tag)
{

  gtk_signal_connect(GTK_OBJECT(tag),
                     "event",
                     GTK_SIGNAL_FUNC(tag_event_handler),
                     NULL);
}

/* these symbols we get from the scanner */
extern const char *get_name (void);
extern const char **get_token_names (void);
extern const char **get_block_names (void);
extern int get_token_no (const char *);
extern int get_block_no (const char *);
extern void token_func (void (*func)(int id, int precedence,
				     int length, int pos, int block_id,
				     int nesting, int block_type));
extern void char_func (int (*func)());
extern int get_state (void);
extern void set_state (int state);
extern int lex ();

/* hilite pattern test */
/* hilite pattern test */
static void
install_pat (GtkEditor *cbuff)
{
  static GdkColor blue, greenish, blueish, red, brownish;
  GtkTextTag *tag;

  if(!g_cscanner) {    		
    g_cscanner = gtk_editor_load_scanner ("c-scanner");
    if(!g_cscanner)
      return;
  }

  gtk_editor_install_scanner (cbuff, g_cscanner);

  blue.pixel = 0;
  blue.red = 0;
  blue.green = 0;
  blue.blue = 65535;

  greenish.pixel = 0;
  greenish.red = 30000;
  greenish.green = 45000;
  greenish.blue = 0;

  blueish.pixel = 0;
  blueish.red = 40000;
  blueish.green = 0;
  blueish.blue = 55000;

  red.pixel = 0;
  red.red = 65535;
  red.green = 0;
  red.blue = 0;

  brownish.pixel = 0;
  brownish.red = 65535;
  brownish.green = 43908;
  brownish.blue = 16383;

  tag = gtk_editor_create_tag(cbuff, "keyword");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground_gdk", &blue,
                 "font", "-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "type");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground_gdk", &greenish,
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "include");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground_gdk", &blueish,
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "pre_processor");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground_gdk", &blueish,
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground_gdk", &red,
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground_gdk", &brownish,
                 NULL);
  setup_tag(tag);
}

static void
new_cbuff (void)
{
  GtkWidget *sw;
  GtkWidget *tkxt;
  GtkWidget *label = gtk_label_new ("C buffer");
  GtkEditor *editor = gtk_editor_new ();
  GtkStyle *style;

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);

  tkxt = gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(editor));
  style = gtk_widget_get_style(GTK_WIDGET(tkxt));
  style->font = gdk_font_load ("-adobe-courier-medium-r-normal-*-12-*-*-*-*-*-*-*");
  gtk_widget_set_style(GTK_WIDGET(tkxt), style);
  gtk_container_add(GTK_CONTAINER(sw), tkxt);

  install_pat (editor);

  gtk_widget_show (sw);
  gtk_widget_show (label);
  gtk_widget_show (tkxt);

  gtk_notebook_append_page (book, sw, label);
}

static void
book_switch_page (GtkNotebook     *nbook,
		  GtkNotebookPage *page)
{
  GList* child = gtk_container_children (GTK_CONTAINER(page->child));

  mainbuffer = GTK_EDITOR(gtk_text_view_get_buffer(GTK_TEXT_VIEW(child->data)));
}

static void
quit (void)
{
  gtk_exit (0);
}

/* main program */
static void
display_greeting (GtkEditor *e)
{
  gtk_text_buffer_insert_at_char (GTK_TEXT_BUFFER(e), 0,
		   "/* Welcome to this little editor test program.\n"
		   " *\n"
		   " * This text should be highlighted as a comment\n"
		   " * and the program below should like nice and coloured.\n"
		   " * If it does not, there's something wrong, and you should\n"
		   " * inform me on email <mailund@daimi.au.dk>\n"
		   " *\n"
		   " * Now, enjoy hacking with the editor widget.\n"
		   " *\n"
		   " *\t/mailund\n"
		   " */\n"
		   "\n"
		   "#include <stdio.h>\n"
		   "\n"
		   "int\n"
		   "main (int argc, char *argv[])\n"
		   "{\n"
		   "  if (argc < 1) {\n"
		   "    printf (\"hello, world!\\n\");\n"
		   "  } else {\n"
		   "    printf (\"howdy, %s\\n\", argv[1]);\n"
		   "  }\n\n"
		   "  return 0;\n"
		   "}\n", -1);
}

static GtkWidget*
create_window ()
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *notebook;
  GtkEditor *editor;
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *hbox;
  GtkWidget *button;
  GtkWidget *sw;
  GtkWidget *tkxt;
  GtkStyle *style;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Editor Test");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  /* -- notebook for buffers -- */
  notebook = gtk_notebook_new ();
  book = GTK_NOTEBOOK (notebook);
  gtk_widget_set_usize (notebook, 450, 250);
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (notebook), "switch_page",
		      GTK_SIGNAL_FUNC (book_switch_page), NULL);

  /* -- first buffer -- */
  mainbuffer = editor = gtk_editor_new();
  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);

  tkxt = gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(editor));
  style = gtk_widget_get_style (tkxt);
  style = gtk_style_copy(style);
  style->font = gdk_font_load ("-adobe-courier-medium-r-normal-*-12-*-*-*-*-*-*-*");
  gtk_widget_set_style (tkxt, style);
  gtk_container_add(GTK_CONTAINER(sw), tkxt);

  install_pat(editor);
  display_greeting(editor);

  gtk_widget_show (sw);
  gtk_widget_show (tkxt);

  label = gtk_label_new ("C buffer");
  gtk_widget_show (label);
  gtk_notebook_append_page (book, sw, label);

  /* -- minibuffer -- */

  entry = gtk_entry_new ();
  gtk_widget_show (entry);
  gtk_box_pack_start (GTK_BOX (vbox), entry, FALSE, FALSE, 0);

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (quit),
		      NULL);

  minibuffer = GTK_ENTRY (entry);
  gtk_signal_connect (GTK_OBJECT (minibuffer), "key_press_event",
		      GTK_SIGNAL_FUNC (minibuf_key_press),
		      NULL);
  mainbuffer = GTK_EDITOR (editor);

  /* -- buttons -- */
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

  button = gtk_button_new_with_label ("New C buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_cbuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Clear");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (clear_buffer), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Load file");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (load_file), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Save file");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (save_file), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  return window;
}

int
main (int argc, char *argv[]) 
{
  GtkWidget *main_window;

  /* initialize */
  gtk_init (&argc, &argv);
  main_window = create_window ();

  /* show */
  gtk_widget_show (main_window);
  gtk_widget_grab_focus (GTK_WIDGET (mainbuffer));

  /* args? */
  if (argc > 1) {
    load_file_cb (g_strdup (argv[1])); /* we need a copy 'cause load
					* frees the string */
  }

  /* main loop */
  gtk_main ();
  g_print("here\n");
  return EXIT_SUCCESS;
}
