/*  Test of the gtk editor widget.
 *  Copyright (C) 1998, 2000 Thomas Mailund <mailund@daimi.au.dk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include "gtkeditor.h"
#include <string.h>

#define BUF_MAX_SIZE 128	/* seems fair for one line */

/* global vars */
static GtkEditor *mainbuffer;
static GtkNotebook *book;
static GtkEntry *minibuffer;
static void (*minibuffer_callback)(char *) = NULL;
static GtkEditorScanner *g_cscanner = NULL;
static GtkEditorScanner *g_javascanner = NULL;
static GtkEditorScanner *g_perlscanner = NULL;
static GtkEditorScanner *g_smlscanner = NULL;
static GtkEditorScanner *g_schemescanner = NULL;
static GtkEditorScanner *htmlscanner = NULL;

/* minibuffer stuff */
static gint
minibuf_key_press (GtkWidget *widget, GdkEventKey *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->keyval == GDK_Return) {
    if (minibuffer_callback) {
      /* The callback is called with a copy of the text. It needs to
       * take care of freeing the text, should that be done. */
      minibuffer_callback (g_strdup(gtk_entry_get_text (minibuffer)));
    }
    gtk_entry_set_text (minibuffer, "");
  }

  return FALSE;
}

static void
get_text_from_minibuffer (GtkSignalFunc callback)
{
  gtk_entry_set_text (minibuffer, "");
  minibuffer_callback = callback;
}

/* callbacks */
static void
load_file_cb (char *filename)
{
  char buf[BUF_MAX_SIZE];
  FILE *fp = fopen (filename, "r");

#ifdef DEBUG
  g_print ("load_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  /* clear old text */
  if (gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(mainbuffer))-1) {
    gtk_text_buffer_delete_chars (GTK_TEXT_BUFFER(mainbuffer), 0, gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER (mainbuffer))-1);
  }

  /* load new */
  do {
    if (fgets (buf, BUF_MAX_SIZE, fp) == NULL) {
      break;
    } else {
      gtk_text_buffer_insert_at_char (GTK_TEXT_BUFFER(mainbuffer),
                                      gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER (mainbuffer))-1, buf, strlen (buf));
    }
  } while (!feof (fp));

  /* clean up */
  fclose (fp);
  g_free (filename);
}

static void
save_file_cb (char *filename)
{
  int len;
  char *buf;
  FILE *fp = fopen (filename, "w");

#ifdef DEBUG
  g_print ("save_file_cb (%s)\n", filename);
#endif

  if (!fp) return;
 
  len = gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER (mainbuffer))-1;
  buf = gtk_text_buffer_get_text_chars (GTK_TEXT_BUFFER (mainbuffer), 0, len, FALSE);
  fputs (buf, fp);
  g_free (buf);

  fclose (fp);
}


static void
clear_buffer (void)
{
  gtk_text_buffer_delete_chars (GTK_TEXT_BUFFER(mainbuffer), 0, gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER (mainbuffer))-1);
}

static void
load_file (void)
{
  get_text_from_minibuffer (load_file_cb);
}

static void
save_file (void)
{
  get_text_from_minibuffer (save_file_cb);
}

static gint
tag_event_handler(GtkTextTag *tag, GtkWidget *widget, GdkEvent *event,
                  const GtkTextIter *iter, gpointer user_data)
{
  gint char_index;

  char_index = gtk_text_iter_get_char_index(iter);

  switch (event->type)
    {
    case GDK_MOTION_NOTIFY:
      printf("Motion event at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_BUTTON_PRESS:
      printf("Button press at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_2BUTTON_PRESS:
      printf("Double click at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_3BUTTON_PRESS:
      printf("Triple click at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_BUTTON_RELEASE:
      printf("Button release at char %d tag `%s'\n",
             char_index, tag->name);
      break;

    case GDK_KEY_PRESS:
    case GDK_KEY_RELEASE:
    case GDK_ENTER_NOTIFY:
    case GDK_LEAVE_NOTIFY:
    case GDK_PROPERTY_NOTIFY:
    case GDK_SELECTION_CLEAR:
    case GDK_SELECTION_REQUEST:
    case GDK_SELECTION_NOTIFY:
    case GDK_PROXIMITY_IN:
    case GDK_PROXIMITY_OUT:
    case GDK_DRAG_ENTER:
    case GDK_DRAG_LEAVE:
    case GDK_DRAG_MOTION:
    case GDK_DRAG_STATUS:
    case GDK_DROP_START:
    case GDK_DROP_FINISHED:
    default:
      break;
    }

  return FALSE;
}

static void
setup_tag(GtkTextTag *tag)
{

  gtk_signal_connect(GTK_OBJECT(tag),
                     "event",
                     GTK_SIGNAL_FUNC(tag_event_handler),
                     NULL);
}

/* hilite pattern test */
static void
install_cpat (GtkEditor *cbuff)
{
  GtkTextTag *tag;
  GtkTextIter start, end;

  if(!g_cscanner) {    		
    g_cscanner = gtk_editor_load_scanner ("c-scanner");
    if(!g_cscanner)
      return;
  }

  gtk_editor_install_scanner (cbuff, g_cscanner);
  
  tag = gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(cbuff), "regular");
  

  tag = gtk_editor_create_tag(cbuff, "keyword");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 "font", "-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "type");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "include");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#9522d8",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "pre_processor");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#9522d8",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#ff0000",
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#cc901a",
                 NULL);
  setup_tag(tag);
    
  tag = gtk_editor_create_tag(cbuff, "escape_char");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);
  

  tag = gtk_editor_create_tag(cbuff, "function");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 NULL);
  setup_tag(tag);
  
  gtk_text_buffer_get_iter_at_char (GTK_TEXT_BUFFER(cbuff),
                                    &start, 0);
  gtk_text_buffer_get_iter_at_char (GTK_TEXT_BUFFER(cbuff),
                                    &end, gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(cbuff)));                          
  gtk_text_buffer_apply_tag (GTK_TEXT_BUFFER(cbuff),
                             "regular", &start, &end);
}

static void
install_perlpat (GtkEditor *perlbuff)
{
  GtkTextTag *tag;
  GtkTextIter start, end;

  if(!g_perlscanner) {    		
    g_perlscanner = gtk_editor_load_scanner ("perl-scanner");
    if(!g_perlscanner)
      return;
  }

  gtk_editor_install_scanner (perlbuff, g_perlscanner);
  
  tag = gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(perlbuff), "regular");
  

  tag = gtk_editor_create_tag(perlbuff, "keyword");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 "font", "-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  
  setup_tag(tag);

  tag = gtk_editor_create_tag(perlbuff, "type");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(perlbuff, "variable");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(perlbuff, "include");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#9522d8",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(perlbuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#ff0000",
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(perlbuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#cc901a",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(perlbuff, "escape_char");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);

  gtk_text_buffer_get_iter_at_char (GTK_TEXT_BUFFER(perlbuff),
                                    &start, 0);
  gtk_text_buffer_get_iter_at_char (GTK_TEXT_BUFFER(perlbuff),
                                    &end, gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(perlbuff)));                          
  gtk_text_buffer_apply_tag (GTK_TEXT_BUFFER(perlbuff),
                             "regular", &start, &end);
}


static void
install_javapat (GtkEditor *javabuff)
{
  GtkTextTag *tag;
  GtkTextIter start, end;

  if(!g_javascanner) {    		
    g_javascanner = gtk_editor_load_scanner ("java-scanner");
    if(!g_javascanner)
      return;
  }

  gtk_editor_install_scanner (javabuff, g_javascanner);

  tag = gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(javabuff), "regular");

  tag = gtk_editor_create_tag(javabuff, "keyword");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 "font", "-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  
  setup_tag(tag);

  tag = gtk_editor_create_tag(javabuff, "type");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(javabuff, "importing");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#3dbfbf",
                 NULL);
  setup_tag(tag);
    
  tag = gtk_editor_create_tag(javabuff, "import");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#9522d8",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(javabuff, "class");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#9522d8",
                 "font", "-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(javabuff, "function");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 NULL);
  setup_tag(tag);


  tag = gtk_editor_create_tag(javabuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#ff0000",
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(javabuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#cc901a",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(javabuff, "escape_char");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);

  
  gtk_text_buffer_get_iter_at_char (GTK_TEXT_BUFFER(javabuff),
                                    &start, 0);
  gtk_text_buffer_get_iter_at_char (GTK_TEXT_BUFFER(javabuff),
                                    &end,
				    gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(javabuff)));
  gtk_text_buffer_apply_tag (GTK_TEXT_BUFFER(javabuff),
                             "regular", &start, &end);
}

static void
install_smlpat (GtkEditor *smlbuff)
{
  GtkTextTag *tag;

  if(!g_smlscanner) {    		
    g_smlscanner = gtk_editor_load_scanner ("sml-scanner");
    if(!g_smlscanner)
      return;
  }

  gtk_editor_install_scanner (smlbuff, g_smlscanner);

  /* pattern properties */
  tag = gtk_editor_create_tag(smlbuff, "keyword");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 "font", "-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(smlbuff, "type");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#4ad822",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(smlbuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#ff0000",
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(smlbuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#cc901a",
                 NULL);
  setup_tag(tag);
}

static void
install_schemepat (GtkEditor *schemebuff)
{
  GtkTextTag *tag;

  if(!g_smlscanner) {    		
    g_schemescanner = gtk_editor_load_scanner ("scheme-scanner");
    if(!g_schemescanner)
      return;
  }

  gtk_editor_install_scanner (schemebuff, g_schemescanner);

  /* pattern properties */
  tag = gtk_editor_create_tag(schemebuff, "keyword");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#9522d8",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(schemebuff, "function");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(schemebuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#ff0000",
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(schemebuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#cc901a",
                 NULL);
  setup_tag(tag);
}

/* hilite pattern test */
static void
install_htmlpat (GtkEditor *cbuff)
{
  GtkTextTag *tag;

  if(!htmlscanner) {    		
    htmlscanner = gtk_editor_load_scanner ("html-scanner");
    if(!htmlscanner)
      return;
  }

  gtk_editor_install_scanner (cbuff, htmlscanner);

  tag = gtk_editor_create_tag(cbuff, "table_tag");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#1dd3c4",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "script_tag");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "#ff0000",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "anchor_tag");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "darkgreen",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "image_tag");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "purple",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "tag");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "darkblue",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "comment");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "grey",
                 "font", "-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*",
                 NULL);
  setup_tag(tag);

  tag = gtk_editor_create_tag(cbuff, "string");
  gtk_object_set(GTK_OBJECT(tag),
                 "foreground", "blue",
                 NULL);
  setup_tag(tag);
}

static GtkEditor *
new_buffer (char *name, void (*pat_func) (GtkEditor *buff))
{
  GtkEditor *editor = gtk_editor_new();
  GtkTextTag *tag = gtk_editor_get_default_tag(editor);
  GtkWidget *vp;
  GtkWidget *sw;
  GtkWidget *tkxt;
  GtkWidget *label = gtk_label_new (name);
  GtkWidget *page;

  /* setup the default tag */
  gtk_object_set(GTK_OBJECT(tag),
                 "pixels_above_lines", 0,
                 "pixels_below_lines", 0,
                 "font", "-adobe-courier-medium-r-normal-*-12-*-*-*-*-*-*-*",
                 NULL);

  vp = gtk_viewport_new (NULL, NULL);

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);
  tkxt = gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(editor));
  gtk_container_add(GTK_CONTAINER(sw), tkxt);
  gtk_container_add(GTK_CONTAINER(vp), sw);

  pat_func (editor);

  gtk_widget_show (vp);
  gtk_widget_show (sw);
  gtk_widget_show (label);
  gtk_widget_show (tkxt);

  gtk_notebook_prepend_page (book, vp, label);

  page = gtk_notebook_get_nth_page (book, 0);
  g_dataset_set_data (page, "editor", editor);

  gtk_notebook_set_page (book, 0);

  return editor;
}

static void
new_cbuff (void)
{
  new_buffer ("C buffer", install_cpat);
}

static void
new_smlbuff (void)
{
  new_buffer ("SML buffer", install_smlpat);
}

static void
new_schemebuff (void)
{
  new_buffer ("Scheme buffer", install_schemepat);
}

static void
new_htmlbuff (void)
{
  new_buffer ("HTML buffer", install_htmlpat);
}

static void
new_javabuff (void)
{
  new_buffer ("Java buffer", install_javapat);
}

static void
new_perlbuff (void)
{
  new_buffer ("Perl buffer", install_perlpat);
}


static void
book_switch_page (GtkNotebook     *nbook,
		  GtkNotebookPage *page)
{
  mainbuffer = g_dataset_get_data (page->child, "editor");
}


static void
quit (void)
{
  gtk_exit (0);
}

/* main program */
static void
display_greeting (GtkEditor *e)
{
  gtk_text_buffer_insert_at_char (GTK_TEXT_BUFFER(e), 0,
		   "/* Welcome to this little editor test program.\n"
		   " *\n"
		   " * This text should be highlighted as a comment\n"
		   " * and the program below should like nice and coloured.\n"
		   " * If it does not, there's something wrong, and you should\n"
		   " * inform me on email <mailund@daimi.au.dk>\n"
		   " *\n"
		   " * Now, enjoy hacking with the editor widget.\n"
		   " *\n"
		   " *\t/mailund\n"
		   " */\n"
		   "\n"
		   "#include <stdio.h>\n"
		   "\n"
		   "int\n"
		   "main (int argc, char *argv[])\n"
		   "{\n"
		   "  if (argc < 1) {\n"
		   "    printf (\"hello, world!\\n\");\n"
		   "  } else {\n"
		   "    printf (\"howdy, %s\\n\", argv[1]);\n"
		   "  }\n\n"
		   "  return 0;\n"
		   "}\n", -1);
}

static void
display_html_greeting (GtkEditor *e)
{
  gtk_text_buffer_insert_at_char (GTK_TEXT_BUFFER(e), 0,
		   "<html>\n"
		   "<head>\n"
		   "  <title> Simple HTML Example </title>\n"
		   "</head>\n"
		   "\n"
		   "<body bgcolor=\"#FFFFFF\">\n"
		   "<h1> Simple HTML Example </h1>\n"
		   "<p>\n\n"
		   "  This is a simple scanner/highlighter for HTML, but it shows\n"
		   "  some of the nice features of\n"
		   "  <a href=\"http://gtkeditor.sourceforge.net\">GtkEditor</a>\'s\n"
		   "  highlighting such as nested block highlighting.\n\n"
		   "<p>\n"
		   "<table width=\"90%\">\n"
		   "<tr>\n"
		   "  <td>\n"
		   "  <i>\n"
		   "     The table tag above shows what you can do with nested\n"
		   "     block highlighting, the string is one such block which\n"
		   "     is nested inside the table block.\n"
		   "  </i>\n"
		   "  </td>\n"
		   "</tr>\n"
		   "</table>\n"
		   "<p>\n"
		   "Another example of nested block highlighting is nested comments:\n"
		   "\n"
		   "<!-- level 0\n"
		   "    <!-- level 1\n"
		   "        <!-- level 2\n"
		   "           ...\n"
		   "         level 2 -->\n"
		   "    level 1 -->\n"
		   "level 0 -->\n"
		   "\n\n"
		   "Blocks can be nested to any number of levels.\n\n"
		   "</body>\n"
		   "</html>\n", -1);
}

static GtkWidget*
create_window ()
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *notebook;
  GtkEditor *editor;
  GtkWidget *entry;
  GtkWidget *hbox;
  GtkWidget *button;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Editor Test");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  /* -- notebook for buffers -- */
  notebook = gtk_notebook_new ();
  book = GTK_NOTEBOOK (notebook);
  gtk_widget_set_usize (notebook, 450, 250);
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (notebook), "switch_page",
		      GTK_SIGNAL_FUNC (book_switch_page), NULL);
		      		      
  /* -- second buffer -- */
  editor = new_buffer ("HTML buffer", install_htmlpat);
  display_html_greeting (editor); 

  /* -- first buffer -- */
  mainbuffer = editor = new_buffer ("C buffer", install_cpat);
  display_greeting(editor);  

  /* -- minibuffer -- */

  entry = gtk_entry_new ();
  gtk_widget_show (entry);
  gtk_box_pack_start (GTK_BOX (vbox), entry, FALSE, FALSE, 0);

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (quit),
		      NULL);

  minibuffer = GTK_ENTRY (entry);
  gtk_signal_connect (GTK_OBJECT (minibuffer), "key_press_event",
		      GTK_SIGNAL_FUNC (minibuf_key_press),
		      NULL);


  /* -- buttons -- */
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

  button = gtk_button_new_with_label ("New C buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_cbuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("New HTML buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_htmlbuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("New SML buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_smlbuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("New Scheme buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_schemebuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("New Java buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_javabuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("New Perl buffer");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (new_perlbuff), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Clear");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (clear_buffer), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Load file");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (load_file), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  button = gtk_button_new_with_label ("Save file");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (save_file), NULL);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);

  return window;
}

int
main (int argc, char *argv[]) 
{
  GtkWidget *main_window;

  /* initialize */
  gtk_init (&argc, &argv);
  main_window = create_window ();

  /* show */
  gtk_widget_show (main_window);

  /* args? */
  if (argc > 1) {
    load_file_cb (g_strdup (argv[1])); /* we need a copy 'cause load
					* frees the string */
  }

  /* main loop */
  gtk_main ();

  return EXIT_SUCCESS;
}
